/****************************************************************************
 * Copyright (c) 2010-2012, EBM WebSourcing - All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/
 
package com.ebmwebsourcing.easycommons.properties;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.junit.Test;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertTrue;
import static junit.framework.Assert.fail;

import com.ebmwebsourcing.easycommons.io.IOHelper;

/**
 * @author Adrien Ruffie - EBM WebSourcing
 */
public class PropertiesHelperTest {

    private static final String PROPA_RSV_VALUE = "value";

    private static final String PROPB_RSV_VALUE = "/" + PROPA_RSV_VALUE + "/value2";

    private static final String PROPC_RSV_VALUE = PROPB_RSV_VALUE + "/" + PROPA_RSV_VALUE;

    private static final String PROPD_RSV_VALUE = "value3/" + PROPC_RSV_VALUE + " "
            + PROPB_RSV_VALUE;

    private static final String PROPE_RSV_VALUE = PROPD_RSV_VALUE + PROPD_RSV_VALUE;

    private static final String PROPL_RSV_VALUE = "value17";

    private static final String PROPM_RSV_VALUE = "value2/" + PROPC_RSV_VALUE + "value4 /value3";

    private static final String PROPN_RSV_VALUE = PROPB_RSV_VALUE + "/" + PROPE_RSV_VALUE;

    private static final String PROPO_RSV_VALUE = PROPD_RSV_VALUE;

    private static final String PROPP_RSV_VALUE = "${test.propP}";

    private static final String PROPX_RSV_VALUE = "value8" + PROPP_RSV_VALUE;

    private static final String PROPY_RSV_VALUE = PROPX_RSV_VALUE + "/value7";

    private static final String PROP_VALUES_WHITESPACEDELIMITER = "java.util.logging.ConsoleHandler org.ow2.petals.commons.logger.PetalsFileHandler org.ow2.petals.commons.logger.TestingFileHandler";

    private static final String PROP_VALUES_COMMADELIMITER = "java.util.logging.ConsoleHandler, org.ow2.petals.commons.logger.PetalsFileHandler ,org.ow2.petals.commons.logger.TestingFileHandler";

    @Test
    public void testResolvePropertiesOK() throws Exception {
        Properties properties = new Properties();
        addCorrectPropData(properties);

        PropertiesHelper.resolveProperties(properties, properties);
        assertEquals(properties.getProperty("test.propA"), PROPA_RSV_VALUE);
        assertEquals(properties.getProperty("test.propB"), PROPB_RSV_VALUE);
        assertEquals(properties.getProperty("test.propC"), PROPC_RSV_VALUE);
        assertEquals(properties.getProperty("test.propD"), PROPD_RSV_VALUE);
        assertEquals(properties.getProperty("test.propE"), PROPE_RSV_VALUE);
    }

    @Test
    public void testResolvePropertiesWithUnresolvableProperties() throws Exception {
        Properties properties = new Properties();
        addUnexistingPropData(properties);

        PropertiesHelper.resolveProperties(properties, properties);
        assertEquals(properties.getProperty("test.propX"), PROPX_RSV_VALUE);
        assertEquals(properties.getProperty("test.propY"), PROPY_RSV_VALUE);
    }

    @Test
    public void testResolvePropertiesWithALoop() {
        Properties properties = new Properties();
        addLoopPropData(properties);

        try {
            PropertiesHelper.resolveProperties(properties, properties);
            fail();
        } catch (PropertiesException pe) {
            assertEquals("There is a place holder loop in the specified properties",
                    pe.getMessage());
            assertTrue(true);
        }
    }

    @Test
    public void testResolvePropertiesWithOtherProperties() throws Exception {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);

        Properties propertiesToResolve = new Properties();
        addCorrectPropData2(propertiesToResolve);

        PropertiesHelper.resolveProperties(propertiesToResolve, propertiesToUse);
        assertEquals(propertiesToResolve.getProperty("test.propL"), PROPL_RSV_VALUE);
        assertEquals(propertiesToResolve.getProperty("test.propM"), PROPM_RSV_VALUE);
        assertEquals(propertiesToResolve.getProperty("test.propN"), PROPN_RSV_VALUE);
        assertEquals(propertiesToResolve.getProperty("test.propO"), PROPO_RSV_VALUE);
    }

    @Test
    public void testResolveMapWithNoExceptionOK() throws Exception {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);

        Map<String, String> map = new HashMap<String, String>();
        addCorrectMapData(map);

        PropertiesHelper.resolveMapWithNoException(map, propertiesToUse);
        assertEquals(map.get("test.propL"), PROPL_RSV_VALUE);
        assertEquals(map.get("test.propM"), PROPM_RSV_VALUE);
        assertEquals(map.get("test.propN"), PROPN_RSV_VALUE);
        assertEquals(map.get("test.propO"), PROPO_RSV_VALUE);
    }

    @Test
    public void testResolveMapWithNoExceptionWithALoop() {
        Properties propertiesToUse = new Properties();
        addLoopPropData(propertiesToUse);

        Map<String, String> map = new HashMap<String, String>();
        addLoopMapData(map);

        PropertiesHelper.resolveMapWithNoException(map, propertiesToUse);
    }

    @Test
    public void testResolveMapOK() throws Exception {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);

        Map<String, String> map = new HashMap<String, String>();
        addCorrectMapData(map);

        PropertiesHelper.resolveMap(map, propertiesToUse);
        assertEquals(map.get("test.propL"), PROPL_RSV_VALUE);
        assertEquals(map.get("test.propM"), PROPM_RSV_VALUE);
        assertEquals(map.get("test.propN"), PROPN_RSV_VALUE);
        assertEquals(map.get("test.propO"), PROPO_RSV_VALUE);
    }

    @Test
    public void testResolveMapWithALoop() throws Exception {
        Properties propertiesToUse = new Properties();
        addLoopPropData(propertiesToUse);

        Map<String, String> map = new HashMap<String, String>();
        addLoopMapData(map);

        try {
            PropertiesHelper.resolveMap(map, propertiesToUse);
            fail();
        } catch (PropertiesException pe) {
            assertEquals("There is a place holder loop in the specified properties",
                    pe.getMessage());
            assertTrue(true);
        }
    }

    @Test
    public void testResolveStringWithNoPlaceHolder() {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);
        addLoopPropData(propertiesToUse);
        addUnexistingPropData(propertiesToUse);

        try {
            String resolvedStr = PropertiesHelper.resolveString("no_place_holder", propertiesToUse);
            assertEquals("no_place_holder", resolvedStr);
        } catch (PropertiesException Pe) {
            fail();
        }
    }

    @Test
    public void testResolveStringWithNoRecursion() throws Exception {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);

        String resolvedStr = PropertiesHelper.resolveString("${test.propA}", propertiesToUse);
        assertEquals(PROPA_RSV_VALUE, resolvedStr);
    }

    @Test
    public void testResolveStringWithARecursion() throws Exception {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);

        String resolvedStr = PropertiesHelper.resolveString("${test.propB}", propertiesToUse);
        assertEquals(PROPB_RSV_VALUE, resolvedStr);
    }

    @Test
    public void testResolveStringWithMoreThanOneRecursion() throws Exception {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);

        String resolvedStr = PropertiesHelper.resolveString("${test.propC}", propertiesToUse);
        assertEquals(PROPC_RSV_VALUE, resolvedStr);
    }

    @Test
    public void testResolveStringWithSeveralRecursions0() throws Exception {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);

        String resolvedStr = PropertiesHelper.resolveString("${test.propD}", propertiesToUse);
        assertEquals(PROPD_RSV_VALUE, resolvedStr);
    }

    @Test
    public void testResolveStringWithSeveralRecursions1() throws Exception {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);

        String resolvedStr = PropertiesHelper.resolveString("${test.propE}", propertiesToUse);
        assertEquals(PROPE_RSV_VALUE, resolvedStr);
    }

    @Test
    public void testResolveStringWithSeveralRecursions2() throws Exception {
        Properties propertiesToUse = new Properties();
        addCorrectPropData(propertiesToUse);

        String resolvedStr = PropertiesHelper.resolveString("${test.propE}${test.propC}",
                propertiesToUse);
        assertEquals(PROPE_RSV_VALUE + PROPC_RSV_VALUE, resolvedStr);
    }

    @Test
    public void testResolveStringAnUnresolvableProperty() throws Exception {
        Properties propertiesToUse = new Properties();
        addUnexistingPropData(propertiesToUse);

        String resolvedStr = PropertiesHelper.resolveString("${test.propX}", propertiesToUse);
        assertEquals(PROPX_RSV_VALUE, resolvedStr);
    }

    @Test
    public void testResolveStringAPropertyLoop() {
        Properties propertiesToUse = new Properties();
        addLoopPropData(propertiesToUse);

        try {
            PropertiesHelper.resolveString("${test.propI}", propertiesToUse);
            fail();
        } catch (PropertiesException pe) {
            assertEquals("There is a place holder loop in the specified properties",
                    pe.getMessage());
            assertTrue(true);
        }
    }

    @Test
    public void testContainsPlaceHolderWithAPlaceHolder() throws Exception {
        assertTrue(PropertiesHelper.containsPlaceHolder("value2/${test.propC}value4 /value3"));
    }

    @Test
    public void testContainsPlaceHolderWithNoPlaceHolder() throws Exception {
        assertTrue(!PropertiesHelper.containsPlaceHolder("value2/value4 /value3"));
    }

    private static final void addCorrectMapData(Map<String, String> map) {
        map.put("test.propL", "value17");
        map.put("test.propM", "value2/${test.propC}value4 /value3");
        map.put("test.propN", "${test.propB}/${test.propE}");
        map.put("test.propO", "${test.propD}");
    }

    private static final void addCorrectPropData(Properties properties) {
        properties.put("test.propA", "value");
        properties.put("test.propB", "/${test.propA}/value2");
        properties.put("test.propC", "${test.propB}/${test.propA}");
        properties.put("test.propD", "value3/${test.propC} ${test.propB}");
        properties.put("test.propE", "${test.propD}${test.propD}");
    }

    private static final void addLoopMapData(Map<String, String> map) {
        map.put("test.propJ", "${test.propI}value5value6");
    }

    private void addCorrectPropData2(Properties propertiesToResolve) {
        propertiesToResolve.put("test.propL", "value17");
        propertiesToResolve.put("test.propM", "value2/${test.propC}value4 /value3");
        propertiesToResolve.put("test.propN", "${test.propB}/${test.propE}");
        propertiesToResolve.put("test.propO", "${test.propD}");
    }

    private static final void addLoopPropData(Properties propertiesToUse) {
        propertiesToUse.put("test.propI", "value4${test.propJ}value5");
        propertiesToUse.put("test.propJ", "${test.propK}value5value6");
        propertiesToUse.put("test.propK", "${test.propJ}");
    }

    private static final void addUnexistingPropData(Properties propertiesToUse) {
        propertiesToUse.put("test.propX", "value8${test.propP}");
        propertiesToUse.put("test.propY", "${test.propX}/value7");
    }

    private static final InputStream addCorrectDataForInputStream() {

        byte[] testByteArray = "test = ${test.propIn}".getBytes();
        return new ByteArrayInputStream(testByteArray);
    }

    private static final InputStream addBadDataForInputStream() {

        byte[] testByteArray = "test".getBytes();
        return new ByteArrayInputStream(testByteArray);
    }

    private static final void addResolvedProperties(Properties propertiesToUse) {
        propertiesToUse.put("test.propIn", "value");
    }

    @Test
    public void testResolvePropertiesForInputStream() throws Exception {
        InputStream inputStreamToResolve = addCorrectDataForInputStream();
        Properties propertiesToUse = new Properties();
        addResolvedProperties(propertiesToUse);

        ByteArrayInputStream inputStreamResolved = null;

        try {
            inputStreamResolved = PropertiesHelper.resolvePropertiesForInputStream(
                    inputStreamToResolve, propertiesToUse);

            Properties propertiesResolved = new Properties();
            propertiesResolved.load(inputStreamResolved);
            assertEquals(propertiesResolved.getProperty("test"), PROPA_RSV_VALUE);

        } finally {
            IOHelper.close(inputStreamToResolve);
            IOHelper.close(inputStreamResolved);
        }
    }

    @Test
    public void testResolvePropertiesWithBadInputStream() throws Exception {
        InputStream inputStreamToResolve = addBadDataForInputStream();
        Properties propertiesToUse = new Properties();
        addResolvedProperties(propertiesToUse);

        ByteArrayInputStream inputStreamResolved = null;

        try {
            inputStreamResolved = PropertiesHelper.resolvePropertiesForInputStream(
                    inputStreamToResolve, propertiesToUse);

            Properties propertiesResolved = new Properties();
            propertiesResolved.load(inputStreamResolved);
            assertEquals(propertiesResolved.getProperty("test"), "");

        } finally {
            IOHelper.close(inputStreamToResolve);
            IOHelper.close(inputStreamResolved);
        }
    }

    @Test
    public void testFlattenPropertiesWithoutDefaultProperties() {
        Properties properties = new Properties();
        properties.setProperty("name1", "value1");
        properties.setProperty("name2", "value2");
        properties.setProperty("name3", "value3");
        properties.setProperty("name4", "value4");
        properties.setProperty("name4", "value5");

        Properties expected = new Properties();
        expected.put("name1", "value1");
        expected.put("name2", "value2");
        expected.put("name3", "value3");
        expected.put("name4", "value5");
        assertEquals(expected, PropertiesHelper.flattenProperties(properties));
    }

    @Test
    public void testFlattenPropertiesWithDefaultProperties() {
        Properties defaultProperties = new Properties();
        Properties properties = new Properties(defaultProperties);

        defaultProperties.setProperty("name1", "value1");
        defaultProperties.setProperty("name2", "value2");
        defaultProperties.setProperty("name4", "value5");

        properties.setProperty("name3", "value3");
        properties.setProperty("name4", "value4");

        Properties expected = new Properties();
        expected.put("name1", "value1");
        expected.put("name2", "value2");
        expected.put("name3", "value3");
        expected.put("name4", "value4");
        assertEquals(expected, PropertiesHelper.flattenProperties(properties));
    }

    @Test
    public void testParsePropertyValues() {
        String[] values = PropertiesHelper.parsePropertyValues(PROP_VALUES_WHITESPACEDELIMITER);
        assertEquals(values[0], "java.util.logging.ConsoleHandler");
        assertEquals(values[1], "org.ow2.petals.commons.logger.PetalsFileHandler");
        assertEquals(values[2], "org.ow2.petals.commons.logger.TestingFileHandler");

        String[] values2 = PropertiesHelper.parsePropertyValues(PROP_VALUES_COMMADELIMITER);
        assertEquals(values2[0], "java.util.logging.ConsoleHandler");
        assertEquals(values2[1], "org.ow2.petals.commons.logger.PetalsFileHandler");
        assertEquals(values2[2], "org.ow2.petals.commons.logger.TestingFileHandler");
    }

}
